/**
 * ==============================================
 *  DQL Adapter for SimQPN
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Simon Eismann, Juergen Walter and other contributors.
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.connector.simqpn.tests;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.dom4j.Element;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.junit.Test;

import tools.descartes.dql.connector.simqpn.ModelStructureQueryConnectorImpl;
import tools.descartes.dql.connector.simqpn.PerformanceMetricsQueryConnectorImpl;
import tools.descartes.dql.connector.simqpn.configuration.ConfigurationBuilder;
import tools.descartes.dql.connector.simqpn.metrics.PerformanceMetric;
import tools.descartes.dql.connector.simqpn.tests.util.TestUtil;
import tools.descartes.dql.connector.simqpn.util.EntityMappingUtil;
import tools.descartes.dql.connector.simqpn.util.FileUtil;
import tools.descartes.dql.connector.simqpn.util.PerformanceMetricsUtil;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.MappingFactory;
import tools.descartes.dql.models.mapping.mapping.impl.MappingFactoryImpl;

public class PerformanceMetricsQueryConnectorImplTest {

		/**
		 * See ModelStructureQueryConnectorImplTest
		 */
		@Test
		public void testAgainstExpectedResults() {
			String modelName = "ispass03-02.qpe";
			String resultName = "expectedPerformanceMetrics_ispass03-02.txt";
			String modelPath = FileUtil.getPath(modelName);
			String resultPath = FileUtil.getPath(resultName);
			
			PerformanceMetricsQueryConnectorImpl pmqc = new PerformanceMetricsQueryConnectorImpl();
			pmqc.setConstraint("Accurate");
			ModelStructureQueryConnectorImpl msqc = new ModelStructureQueryConnectorImpl();
			
			//Build request
			MappingFactory mappingFactory = MappingFactoryImpl.init();
			EntityMapping request = mappingFactory.createEntityMapping();
			request.setModelLocation(modelPath);
			
			//Run request
			request = msqc.requestEntities(request);
			request = msqc.requestProbes(request);
		TestUtil.addDefaultResultType(request);
			EntityMapping answer = pmqc.requestMetrics(request);
			
			//Load allowed Ranges
			Object[][] allowedRanges = TestUtil.loadArrayFromFile(resultPath);
			
			List<String> violations = new LinkedList<String>();
			for (int i = 0; i < allowedRanges.length; i++) {
				Double result = TestUtil.getValueOf(answer, (String)allowedRanges[i][0],
						(String)allowedRanges[i][1]);
				if (result < (Double)allowedRanges[i][2] ||
						result > (Double)allowedRanges[i][3])
					violations.add("The value " + result + " of " + (String)allowedRanges[i][0] 
							+ "." + (String)allowedRanges[i][1] + " is outside of the expected Range [" 
							+ (Double)allowedRanges[i][2] + "," + (Double)allowedRanges[i][3] + "]");
			}
			
			if (!violations.isEmpty())
				fail(TestUtil.listToString(violations));
		}
		
		/*
		 * This test looks complicated, but is actually quite simple.
		 * It requests different metrics for the only place in the net and
		 * then checks if the statlevel were set accordingly.
		 * In a second phase it requests 2 metrics for the only place in the net
		 * and checks if the stat level was set accordingly (it should be the higher one 
		 * of the two metrics)
		 */
		@Test
		public void setAllConfigLevelsViaMetrics() {
			String modelName = "SingleQueue.qpe";
			List<PerformanceMetric> performanceMetrics = new ArrayList<PerformanceMetric>();
			List<String> names = new LinkedList<String>();
			names.add("color");
			performanceMetrics.addAll(PerformanceMetricsUtil.getPerformanceMetricsForQueueingPlace(names, "queue"));
		    
			ModelStructureQueryConnectorImpl msqc = new ModelStructureQueryConnectorImpl();
			PerformanceMetricsQueryConnectorImpl pmqc = new PerformanceMetricsQueryConnectorImpl();
			
			//get model
			String modelPath = FileUtil.getPath(modelName);
			Element net = FileUtil.loadXmlFromFile(modelPath);
			
			//Setup basic request
			MappingFactory mappingFactory = MappingFactoryImpl.init();
			EntityMapping request = mappingFactory.createEntityMapping();
			request.setModelLocation(modelPath);
			request = msqc.requestEntities(request);
			
			//Iterate over all availible Metrics
			for (PerformanceMetric pm: performanceMetrics) {
				ConfigurationBuilder cb = new ConfigurationBuilder(net);
				EntityMapping requestWithMetric = EcoreUtil.copy(request);
			requestWithMetric.getEntities().get(0).getProbes().add(
						EntityMappingUtil.createProbe(pm.getDQLName()));
				pmqc.setAllConfigLevelsViaMetrics(cb, net, requestWithMetric);
				if (cb.getStatLevels().get(0).getLevel() != pm.getStatLevel())
					fail("For the Metric " + pm.getDQLName() + " the StatLevel was: " 
				+ cb.getStatLevels().get(0).getLevel() + " expected: " + pm.getStatLevel());
			}
			
			//Metric combinations because we can
			for (PerformanceMetric metric: performanceMetrics) {
				for (PerformanceMetric metric2: performanceMetrics) {
					ConfigurationBuilder cb = new ConfigurationBuilder(net);
					EntityMapping requestWithMetric = EcoreUtil.copy(request);
				requestWithMetric.getEntities().get(0).getProbes().add(
							EntityMappingUtil.createProbe(metric.getDQLName()));
				requestWithMetric.getEntities().get(0).getProbes()
						.add(
							EntityMappingUtil.createProbe(metric2.getDQLName()));
					pmqc.setAllConfigLevelsViaMetrics(cb, net, requestWithMetric);
					
					int expected = metric.getStatLevel();
					if (metric2.getStatLevel() > expected)
						expected = metric2.getStatLevel();
					
					if (cb.getStatLevels().get(0).getLevel() != expected)
						fail("For the Metrics " + metric.getDQLName() + " and " + metric2.getStatLevel() 
								+ " the StatLevel was: " + cb.getStatLevels().get(0).getLevel() 
								+ " expected: " + expected);
				}
			}
		}

	/*
	 * This simple test checks if the Settings are set correctly.
	 * It uses a new ConfigurationBuilder, then sets the settings
	 * via the PerformanceMetricsQueryConnector (which has a state)
	 */
		
	@Test
	public void setConfigSettingsViaConstraintTest() {
		String[] constraints = new String[] {"Accurate", null, "Fast"};
		Double[] expectedAccuracy = new Double[] {0.95, 0.95, 0.8};

		String modelName = "SingleQueue.qpe";
		String modelPath = FileUtil.getPath(modelName);
		Element net = FileUtil.loadXmlFromFile(modelPath);
		
		for (int i = 0; i < constraints.length; i++) {
			PerformanceMetricsQueryConnectorImpl pmqc = new PerformanceMetricsQueryConnectorImpl();
			pmqc.setConstraint(constraints[i]);
			ConfigurationBuilder cb = new ConfigurationBuilder(net);
			pmqc.setConfigSettingsViaConstraint(cb);
		assertEquals(expectedAccuracy[i], cb.getAccuracy(), 0.0000001);
		}
	}
}
