/**
 * ==============================================
 *  DQL Adapter for SimQPN
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Simon Eismann, Juergen Walter and other contributors.
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.connector.simqpn;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;
import org.dom4j.Element;
import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.dql.connector.ModelStructureQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.connector.simqpn.exception.DQLException;
import tools.descartes.dql.connector.simqpn.metrics.PerformanceMetric;
import tools.descartes.dql.connector.simqpn.setup.Activator;
import tools.descartes.dql.connector.simqpn.util.EntityMappingUtil;
import tools.descartes.dql.connector.simqpn.util.FileUtil;
import tools.descartes.dql.connector.simqpn.util.NetUtil;
import tools.descartes.dql.connector.simqpn.util.PerformanceMetricsUtil;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class ModelStructureQueryConnectorImpl implements ModelStructureQueryConnector {

	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	
	public String getBundleProperty(String key) {
		return Activator.getContext().getProperty(key);
	}
	
	@Override
	public EntityMapping requestDoF(EntityMapping request) {
		Element net = FileUtil.loadXmlFromFile(request.getModelLocation());
		final EntityMapping response = EcoreUtil.copy(request);
		
		List<DoF> dofs = getAvailibleDoFOptions(net);
		response.getDoF().addAll(dofs);
		return response;
	}
	
	@Override
	public EntityMapping requestEntities(EntityMapping request) {
		final EntityMapping response = EcoreUtil.copy(request);
		
		//Load 
		Element net = FileUtil.loadXmlFromFile(request.getModelLocation());
		
		//Add resources
		List<String> resourceNames = getPlaceNames(net);
		resourceNames.addAll(getQueueNames(net));
		response.getEntities().addAll(EntityMappingUtil.createEntity(getProbeNames(net)));
		response.getEntities().addAll(EntityMappingUtil.createEntity(resourceNames));
		
		return response;
	}

	private List<String> getQueueNames(Element net) {
		List<String> queueNames = new LinkedList<String>();
		for (Element e: NetUtil.listQueues(net))
			queueNames.add("Queue_" + NetUtil.getNameOfElement(e));
		return queueNames;
	}

	@Override
	public EntityMapping requestProbes(EntityMapping request) {
		final EntityMapping response = EcoreUtil.copy(request);
		
		//Load 
		Element net = FileUtil.loadXmlFromFile(request.getModelLocation());
		for (Entity r : response.getEntities()) {
			r.getProbes().clear();
			if (!r.getIdentifier().contains("Queue_"))
				r.getProbes().addAll(getAvailibleProbes(NetUtil.getPlaceByName(r.getIdentifier(),
						net), net));
			else 
				r.getProbes().addAll(getAvailibleProbesForQueue(NetUtil.removeQueueTag(r.getIdentifier()),
						net));
		}
			
		return response;
	}

	private Collection<? extends Probe> getAvailibleProbesForQueue(
			String queueName, Element net) {
		List<PerformanceMetric> pms = PerformanceMetricsUtil.getPerformanceMetricsForQueue();

		List<Probe> probes = new LinkedList<Probe>();
		for (PerformanceMetric pm: pms) {
			probes.add(EntityMappingUtil.createProbe(pm.getDQLName()));
		}
		
		return probes;
	}

	@Override
	public boolean supportsDoFRequests() {
		return true;
	}

	@Override
	public boolean supportsEntityRequests() {
		return true;
	}

	@Override
	public boolean supportsProbeRequests() {
		return true;
	}
	
	private Collection<? extends Probe> getAvailibleProbes(Element element, Element net) {
		Collection<Probe> c = new LinkedList<Probe>();
		switch (NetUtil.getTypeOfPlace(element)) {
		case "queueing-place":
			c.addAll(getAvailibleProbesForQueue(element, net));
			break;
		case "ordinary-place":
			c.addAll(getAvailibleProbesForOrdinaryPlace(element, net));
			break;
		case "subnet-place":
			log.error("Metrics for Subnets are not yet implemented!");
			throw new DQLException("The SimQPN Connector currently does not support subnets");
		default:
			break;
		}
		return c;
	}

	private static Collection<? extends Probe> getAvailibleProbesForQueue(
			Element element, Element net) {
		List<String> colorNames = getColorNames(element, net);
		String queueName = NetUtil.getNameOfUnderlyingQueue(
				NetUtil.getNameOfElement(element), net);
		List<PerformanceMetric> metrics = PerformanceMetricsUtil.getPerformanceMetricsForQueueingPlace(
				colorNames, queueName);
		
		List<Probe> probes = new LinkedList<Probe>();
		for (PerformanceMetric pm: metrics) {
			probes.add(EntityMappingUtil.createProbe(pm.getDQLName()));
		}
		
		return probes;
	}

	private static String qualifyByRemoval(String colorName) {
		String ret = "";
		for (int i = 0; i < colorName.length(); i++) {
			String digit = colorName.substring(i, i + 1);
			if (digit.matches("\\w")) {
				ret += digit;
			}
		}
		return ret;
	}
	
	private static List<String> getColorNames(Element place, Element net) {
		List<String> colorNames = new LinkedList<String>();
		for (Element colorRef: NetUtil.listColorReferences(place)) {
			colorNames.add(qualifyByRemoval(NetUtil.getNameOfElement(NetUtil.getElement(net,
					NetUtil.getReferencedColorId(colorRef)))));
		}
		return colorNames;
	}
	
	private static Collection<Probe> getAvailibleProbesForProbe(Element element, Element net) {
		List<String> colorNames = getColorNames(element, net);
		List<PerformanceMetric> metrics = PerformanceMetricsUtil.getPerformanceMetricsForProbe(colorNames);

		List<Probe> probes = new LinkedList<Probe>();
		for (PerformanceMetric pm: metrics) {
			probes.add(EntityMappingUtil.createProbe(pm.getDQLName()));
		}
		
		return probes;
	}
	
	private static Collection<Probe> getAvailibleProbesForOrdinaryPlace(Element element, Element net) {
		List<String> colorNames = getColorNames(element, net);
		List<PerformanceMetric> metrics = PerformanceMetricsUtil.getPerformanceMetricsForPlace(colorNames);
		
		List<Probe> probes = new LinkedList<Probe>();
		for (PerformanceMetric pm: metrics) {
			probes.add(EntityMappingUtil.createProbe(pm.getDQLName()));
		}
		
		return probes;
	}
	
	private static List<DoF> getAvailibleDoFOptions(Element net) {
		List<DoF> dofs = new LinkedList<DoF>();
		for (Element place : NetUtil.listAllPlaces(net)) {
			for (Element colorRef: NetUtil.listColorReferences(place)) {
				String dofIdentifier = createDofIdentifier(
						NetUtil.getNameOfElement(place), 
						NetUtil.getNameOfElement(NetUtil.getElement(net, 
								NetUtil.getReferencedColorId(colorRef))));
				int initialValue = NetUtil.getInitialPopulation(colorRef);
				dofs.add(EntityMappingUtil.createDof(dofIdentifier, initialValue));
			}
				
		}
		return dofs;
	}
	
	private static String createDofIdentifier(String placeName, String colorName) {
		return placeName + "_" + colorName;
	}
	
	private static List<String> getProbeNames(Element net) {
		List<String> probeNames = new LinkedList<String>();
		for (Element e: NetUtil.listProbes(net))
			probeNames.add(NetUtil.getNameOfElement(e));
		return probeNames;
	}
	
	private static List<String> getPlaceNames(Element net) {
		List<String> placeNames = new LinkedList<String>();
		for (Element e: NetUtil.listPlaces(net))
			placeNames.add(NetUtil.getNameOfElement(e));
		return placeNames;
	}

	@Override
	public void addQueryProgressListener(QueryProgressListener arg0) {
		// TODO Auto-generated method stub
		
	}
}
