/**
 * ==============================================
 *  DQL Adapter for SimQPN
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Simon Eismann, Juergen Walter and other contributors.
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.connector.simqpn;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.dom4j.Element;
import org.dom4j.QName;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.w3c.dom.Document;

import de.tud.cs.simqpn.kernel.console.SimQPN;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.connector.simqpn.configuration.ConfigurationBuilder;
import tools.descartes.dql.connector.simqpn.exception.DQLException;
import tools.descartes.dql.connector.simqpn.metrics.PerformanceMetric;
import tools.descartes.dql.connector.simqpn.progress.DQLListenerNotifier;
import tools.descartes.dql.connector.simqpn.progress.SimQPNProgressListener;
import tools.descartes.dql.connector.simqpn.setup.Activator;
import tools.descartes.dql.connector.simqpn.util.EntityMappingUtil;
import tools.descartes.dql.connector.simqpn.util.FileUtil;
import tools.descartes.dql.connector.simqpn.util.NetUtil;
import tools.descartes.dql.connector.simqpn.util.PerformanceMetricsUtil;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.StatType;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class PerformanceMetricsQueryConnectorImpl implements
		PerformanceMetricsQueryConnector {
	
	private static String resultFile = "\\result.txt";
	private static String logFileName = "\\log.txt";
	private static String logPropertiesFile = "\\log_properties.log4j";
	private static String modifiedNetFile = "\\modifiedNet.qpe";
	private static double UNSET_ACCURACY = 0.95;
	private static String ACCURATE = "Accurate";
	private static double ACCURATE_ACCURACY = 0.95;
	private static String FAST = "Fast";
	private static double FAST_ACCURACY = 0.8;
	
	private String pathForDoFResults = null;
	private int[] dofChoices;
	private DQLListenerNotifier listenerNotifier = new DQLListenerNotifier();
	
	private double accuracy = UNSET_ACCURACY;
	private static final Logger log = DQLLogger.getLogger(PerformanceMetricsQueryConnectorImpl.class.getName());
	
	@Override
	public EntityMapping requestMetrics(EntityMapping request) {
		Element net = FileUtil.loadXmlFromFile(request.getModelLocation());
		return simulate(net, request);
	}

	@Override 
	public List<EntityMapping> requestDoFMetrics(EntityMapping request) {
		List<EntityMapping> results = new LinkedList<EntityMapping>();
		List<DegreeOfFreedom> dofs = new LinkedList<DegreeOfFreedom>();
		
		for (DoF dof : request.getDoF())
			dofs.add(parseDegreeOfFreedom(dof));
		
		listenerNotifier.setNumberOfRuns(getNumberOfRuns(dofs));
		
		pathForDoFResults = new File(request.getModelLocation()).getParent() + "\\DQL\\" 
			+ FileUtil.getNameOfFile(request.getModelLocation()) +	"\\" + getCurrentDate() +
			" DoF";
		
		recursiveDoFTraverse(dofs, results, request);
		return results;
	}
	
	private int getNumberOfRuns(List<DegreeOfFreedom> dofs) {
		int numberOfRuns = 1;
		for (DegreeOfFreedom dof: dofs)
			numberOfRuns *= dof.getValues().length;
		return numberOfRuns;
	}

	private void recursiveDoFTraverse(List<DegreeOfFreedom> dofs,
			List<EntityMapping> results, EntityMapping request) {
		Element net = FileUtil.loadXmlFromFile(request.getModelLocation());
		recursiveDoFTraverse(dofs, results, 0, request, net, new int[dofs.size()]);
	}

	private void recursiveDoFTraverse(List<DegreeOfFreedom> dofs, List<EntityMapping> results, 
			int i, EntityMapping request, Element net, int[] choices) {
		if (i >= dofs.size()) {
			this.dofChoices = choices;
			results.add(simulate(net, request));
			return;
		}
		
		for (int valueNumber = 0; valueNumber < dofs.get(i).getValues().length; valueNumber++) {
			choices[i] = dofs.get(i).values[valueNumber];
			EntityMapping newRequest = EcoreUtil.copy(request);
			dofs.get(i).apply(net, valueNumber);
			addDoFSettings(dofs.get(i), valueNumber, newRequest);
			recursiveDoFTraverse(dofs, results, i + 1, newRequest, net, choices);
		}	
	}

	private void addDoFSettings(DegreeOfFreedom degreeOfFreedom, int valueNumber, EntityMapping response) {
		for (DoF dof : response.getDoF())
			if (dof.getIdentifier().matches(degreeOfFreedom.getDQLName())) {
				EList<Probe> doFSettings = dof.getProbes();
				Probe p = DomainFactory.eINSTANCE.createProbe();
				StatType st = DomainFactory.eINSTANCE.createStatType();
				DecimalResult doFSetting = EntityMappingUtil.createDoFParameter(
						degreeOfFreedom.toString(), (double)degreeOfFreedom.getValues()[valueNumber]);

				doFSettings.add(p);
				p.getStatTypes().add(st);
				st.setResult(doFSetting);

				return;
			}
		
	}

	/*
	 * This is encapsulated in a method, since both requstMetrics and requestDoFMetrics use it
	 * Can't use requestMetrics within requestDoFMetrics, since that would require saving the 
	 * modified net twice.
	 */
	private EntityMapping simulate(Element net, EntityMapping request) {
		final EntityMapping response = EcoreUtil.copy(request);
		ConfigurationBuilder cb = new ConfigurationBuilder(net);

		listenerNotifier.updateNextRunStarted();
		
		setConfigSettingsViaConstraint(cb);
		setAllConfigLevelsViaMetrics(cb, net, response);
		
		File resultFolder = runSimulation(cb, FileUtil.getModelPath(request.getModelLocation()), net);
		File results = new File(resultFolder.getAbsolutePath() + resultFile);
		File logFile = new File(resultFolder.getAbsolutePath() + logFileName);
		Document d = FileUtil.loadDocument(results);
		boolean warning = checkForWarning(logFile);
		parseResults(d, response, warning, net);
		response.setDoFModelLocation(FileUtil.getURL(resultFolder) + modifiedNetFile);
		return response;
	}
	
	private boolean checkForWarning(File logFile) {
		try {
			String toCheck = new String(Files.readAllBytes(Paths.get(FileUtil.getURL(logFile))));
			return toCheck.contains("The required precision may not have been reached!");
		} catch (IOException e) {
			log.error("Could not read config File!");
			new DQLException("Internal Error: Could not read config File!");
		}
		return true;
	}

	private void parseResults(Document d, EntityMapping response, boolean warning, Element net) {
		for (Entity r : response.getEntities()) {
			List<Probe> toAdd = new LinkedList<Probe>();
			String typeOfPlace = "queue";
			if(!r.getIdentifier().startsWith("Queue_"))
				typeOfPlace = NetUtil.getTypeOfPlace(NetUtil.getPlaceByName(r.getIdentifier(), net));
			
			for (Iterator<Probe> iter = r.getProbes().iterator(); iter.hasNext(); ) {
				Probe p = iter.next();
				PerformanceMetric pm = PerformanceMetricsUtil.parsePerformanceMetric(
						p.getMetricName(), typeOfPlace);
				Double result = pm.getValue(d, r.getIdentifier(), typeOfPlace);
				
				if (result == null)
					p.getStatTypes().get(0)
							.setResult(EntityMappingUtil.createDecimalResult(p.getMetricName(), -1.0, accuracy, false));
				else
					p.getStatTypes().get(0).setResult(
							EntityMappingUtil.createDecimalResult(p.getMetricName(),
							result, accuracy, !warning));
			}
			
			r.getProbes().addAll(toAdd);
		}
	}
	
	private File runSimulation(ConfigurationBuilder configurationBuilder,
			String inputFilePath, Element net) {
		configurationBuilder.writeConfiguration(net);
		
		//create folder for the Files
		File input = new File(inputFilePath);
		File outputFolder = null;
		
		if (pathForDoFResults == null) {
			outputFolder = createOutputFolder(inputFilePath, input);
		} else {
			outputFolder = new File(pathForDoFResults + "\\" + getCurrentDofChoice());
		}

		outputFolder.mkdirs();
		outputFolder.mkdir();
		
		//Create files
		File results = FileUtil.createFile(outputFolder.getAbsolutePath() + resultFile);
		File logfile = FileUtil.createFile(outputFolder.getAbsolutePath() + logFileName);
		File logProperties = FileUtil.createFile(outputFolder.getAbsolutePath() + logPropertiesFile);
		savePropertiesToFile(logfile, logProperties);
		File modifiedNet = FileUtil.createFile(outputFolder.getAbsolutePath() + modifiedNetFile);
		NetUtil.save(modifiedNet, net);
		SimQPN.runSimulator(FileUtil.getURL(modifiedNet), ConfigurationBuilder.CONFIGNAME, FileUtil.getURL(results),
				FileUtil.getURL(logProperties), new SimQPNProgressListener(listenerNotifier));
		logProperties.delete();
		
		BasicConfigurator.configure();
		log.info("Information about the simulation was written to: " + FileUtil.getURL(results));
		
		return outputFolder;
	}

	private String getCurrentDofChoice() {
		String ret = "";
		for (int i = 0; i < dofChoices.length; i++) {
			ret += dofChoices[i];
			if (i != dofChoices.length - 1)
				ret += " ";
		}
		
		return ret;
	}

	private String getCurrentDate() {
	    Date dNow = new Date( );
	    SimpleDateFormat ft = new SimpleDateFormat ("yyyy MM dd HH mm ss SS");
	    return ft.format(dNow);
	}
	
	private File createOutputFolder(String inputFilePath, File input) {
		File outputFolder = new File(input.getParent() + "\\DQL\\" + FileUtil.getNameOfFile(inputFilePath) +
				"\\" + getCurrentDate());
		int i = 1;
		while (outputFolder.exists()) {
			outputFolder = new File(input.getParent() + "\\DQL\\"  + FileUtil.getNameOfFile(inputFilePath) +
					"\\"+ getCurrentDate() + "_" + i);
			i++;
		}
		return outputFolder;
	}
	
	private void savePropertiesToFile(File log, File propertiesFile) {
	    try {
			Properties props = new Properties();
			props.setProperty("log4j.rootLogger", "WARN, File");
			props.setProperty("log4j.appender.File", "org.apache.log4j.RollingFileAppender");
			props.setProperty("log4j.appender.File.File", FileUtil.getURL(log));
			props.setProperty("log4j.appender.File.layout", "org.apache.log4j.PatternLayout");
			props.setProperty("log4j.appender.File.maxFileSize", "1MB");
	        OutputStream out = new FileOutputStream(propertiesFile);
	        props.storeToXML(out, "This is an optional header comment string");
	    }
	    catch (Exception e ) {
	    	throw new DQLException("Couldn't save properties to file!");
	    }
	}
	
	/*
	 * Public so Junit can acess it, making for an easier testing
	 */
	public void setAllConfigLevelsViaMetrics(ConfigurationBuilder cb, Element net,
			EntityMapping request) {
		
		for (Entity r: request.getEntities())
			for (Probe p: r.getProbes()) {
				if (!r.getIdentifier().startsWith("Queue_"))				
					cb.addStatLevel(NetUtil.getIdOfPlaceByName(r.getIdentifier(), net), 
						PerformanceMetricsUtil.parsePerformanceMetric(
								p.getMetricName(), NetUtil.getTypeOfPlace(
										NetUtil.getPlaceByName(r.getIdentifier(), net))
								).getStatLevel());
				else
					cb.addStatLevel(NetUtil.getIdOfPlaceWithQueue(NetUtil.removeQueueTag(r.getIdentifier()), net),
							PerformanceMetricsUtil.parsePerformanceMetric(p.getMetricName(), "queue").getStatLevel());

			}
	}

	/*
	 * Public so Junit can acess it, making for a easier testing
	 */
	public void setConfigSettingsViaConstraint(ConfigurationBuilder cb) {
		cb.setAccuracy(accuracy);
	}
	
	public DegreeOfFreedom parseDegreeOfFreedom(DoF dof) {
		int[] values = parseDoFString((String)dof.getProperties().get("value"));
		String[] splits = dof.getIdentifier().split("_");
		if (splits.length != 2) {
			log.error("Malformatted DoF identifier: " + dof.getIdentifier());
			throw new DQLException("Malformatted DoF identifier: " + dof.getIdentifier());
		}
		return new DegreeOfFreedom(values, splits[0], splits[1]);
	}
	
	private static int[] parseDoFString(String s) {
		s = s.replaceAll("(\\[|\\])", "");
		s = s.replaceAll("\\s", "");
		String[] stringPieces = s.split(",");
		int[] ret = new int[stringPieces.length];
		for (int i = 0; i < stringPieces.length; i++) 
			try {
				ret[i] = Integer.parseInt(stringPieces[i]);
			} catch (NumberFormatException e) {
				log.error("Could not parse DoF value #" + i + ": " + stringPieces[i]);
				throw new DQLException("The DoF parameters have to be numbers, but "+ stringPieces[i] 
						+ "is not");
			}
		return ret;
	}

	@Override
	public String getBundleProperty(String key) {
		return Activator.getContext().getProperty(key);
	}


	@Override
	public boolean supportsConstraintRequests() {
		return true;
	}

	@Override
	public boolean supportsDoFRequests() {
		return true;
	}

	@Override
	public boolean supportsMetricRequests() {
		return true;
	}

	@Override
	public boolean supportsObservationPeriods() {
		return false;
	}
	
	@Override
	public String getConstraint() {
		return null;
	}

	@Override
	public void reset() {
		this.accuracy = UNSET_ACCURACY;
		this.pathForDoFResults = null;
	}

	@Override
	public void setConstraint(String constraint) {
		this.accuracy = parseConstraint(constraint);
	}

	@Override
	public void addQueryProgressListener(QueryProgressListener listener) {
		listenerNotifier.addListener(listener);
	}
	
	/*
	 * Useless stubs
	 */

	@Override
	public String getDoFConstraint() {
		return null;
	}

	@Override
	public String getDoFExplorationStrategy() {
		return null;
	}

	@Override
	public Properties getDoFExplorationStrategyProperties() {
		return null;
	}

	@Override
	public List<String> getSupportedConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedDoFConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedExplorationStrategyNames() {
		return null;
	}

	@Override
	public List<String> getSupportedTimeUnits() {
		return null;
	}

	@Override
	public void setDoFConstraint(String doFConstraint) {
		
	}

	@Override
	public void setDoFExplorationStrategy(String strategy, Properties properties) {
		
	}

	@Override
	public void setObservationPeriod(Date start, Date end, long sampleSize) {
		
	}

	@Override
	public void setObservationPeriod(long amount, String unit, long sampleSize) {
		
	}

	@Override
	public void setObservationReference(String type, String identifier,
			long sampleSize) {
	}
	
	public static double parseConstraint(String constraint) {
		if (constraint == null)
			return UNSET_ACCURACY;
		if (constraint.equals(ACCURATE))
			return ACCURATE_ACCURACY;
		if (constraint.equals(FAST))
			return FAST_ACCURACY;
		Pattern pattern = Pattern.compile("[Aa]ccuracy[ ]?=[ ]?([0-9]+.?[0-9]+)");
		Matcher m = pattern.matcher(constraint);
		if (m.matches()) {
			double accuracy = Double.parseDouble(m.group(1));
			if (accuracy > 1.0)
				throw new DQLException("Accuracy has to be below 1.0");
			return accuracy;
		}
		PerformanceMetricsQueryConnectorImpl.log.error("Illegal Constraint");
		throw new DQLException(constraint + " is not a valid Constraint!");
	}
	
	private class DegreeOfFreedom {
		private int[] values;
		private String placeName;
		private String colorName;
		
		public DegreeOfFreedom(int[] values, String placeName, String colorName) {
			this.values = values;
			this.placeName = placeName;
			this.colorName = colorName;
		}
		
		public int[] getValues() {
			return values;
		}

		public String toString() {
			return "Initial Population of " + colorName + " in " + placeName;
		}
		
		public void apply(Element net, int valueNumber) {
			Element place = NetUtil.getPlaceByName(placeName, net);
			for (Element colorRef: NetUtil.listColorReferences(place)) {
						if(NetUtil.getNameOfElement(NetUtil.getElement(net, 
								NetUtil.getReferencedColorId(colorRef))).equals(colorName))
							colorRef.addAttribute(new QName("initial-population"), "" + values[valueNumber]);
			}
		}
		
		public String getDQLName() {
			return placeName + "_" + colorName;
		}
	}
}
