/**
 * ==============================================
 *  DQL Adapter for SimQPN
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Simon Eismann, Juergen Walter and other contributors.
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.connector.simqpn.configuration;

import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;
import org.dom4j.Element;

import tools.descartes.dql.connector.simqpn.util.NetUtil;
import tools.descartes.dql.core.engine.util.DQLLogger;

public class ConfigurationBuilder {

	public static final String CONFIGNAME = "dql-config";
	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	private Double warmupPeriod = 50000000.0;;
	private Double totalRunLength = Double.MAX_VALUE;;
	private String stoppingRule = "RELPRC";
	private Double timeBetweenStopChecks = 10.0;;
	private Double secondsBetweenStopChecks = 3.0;;
	private Double accuracy;
	private int verbosity = 1;;
	private String outputDirectory = ".";;
	private int timeBeforeInitialHeartBeat;
	private int secondsBetweenHeartBeats;
	private List<StatLevel> statLevels = new LinkedList<StatLevel>();
	
	public ConfigurationBuilder(Element net) {
		for (Element place: NetUtil.listAllPlaces(net))
			statLevels.add(new StatLevel(NetUtil.getIdOfElement(place), 0));
		for (Element probe: NetUtil.listProbes(net))
			statLevels.add(new StatLevel(NetUtil.getIdOfElement(probe), 0));
	}
	
	public void addStatLevel(String placeId, int level) {
		for (StatLevel sl: statLevels)
			if (placeId.equals(sl.getPlaceId())) {
				if (level > sl.getLevel()) {
					sl.setLevel(level);
				}
				return;
			}
		statLevels.add(new StatLevel(placeId, level));
	}
	
	public void writeConfiguration(Element net) {
		IdManager idManager = new IdManager(net);
		addMainConfig(net, idManager);
		for (StatLevel statLevel: statLevels)
			addStatLevel(statLevel, net,idManager);
	}
	
	public List<StatLevel> getStatLevels() {
		return statLevels;
	}
	
	private Element setupMetaAttributes(Element element) {
		Element metaAttributes = (Element)element.selectSingleNode("meta-attributes");
		if (metaAttributes == null) {
			metaAttributes = element.addElement("meta-attributes");
		}

		if (element.selectSingleNode("meta-attributes/meta-attribute[@configuration-name=\"" + CONFIGNAME + "\"]") != null)
			element.selectSingleNode("meta-attributes/meta-attribute[@configuration-name=\"" + CONFIGNAME + "\"]").detach();
		
		return metaAttributes;
	}
	
	private void addStatLevel(StatLevel statLevel, Element net,IdManager idManager) {
		Element placeElement = NetUtil.getElement(net, statLevel.getPlaceId());
		
		Element metaAttributes = setupMetaAttributes(placeElement);
		
		if (NetUtil.getTypeOfPlace(placeElement)  != "probe")
			metaAttributes.addElement("meta-attribute")
			.addAttribute("xsi:type", "simqpn-place-configuration")
			.addAttribute("configuration-name", CONFIGNAME)
			.addAttribute("id", idManager.getNextId())
			.addAttribute("statsLevel", "" + statLevel.getLevel());
		else
			metaAttributes.addElement("meta-attribute")
			.addAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
			.addAttribute("xsi:type", "simqpn-place-configuration")
			.addAttribute("configuration-name", CONFIGNAME)
			.addAttribute("id", idManager.getNextId())
			.addAttribute("statsLevel", "" + statLevel.getLevel());
		
		
		if (statLevel.getLevel() >= 3) {
			for (Element colorRef: NetUtil.getColorReferencesForPlace(placeElement)) {
				Element refMetaAttributes = setupMetaAttributes(colorRef);
				refMetaAttributes.addElement("meta-attribute")
					.addAttribute("xsi:type", "simqpn-batch-means-color-configuration")
					.addAttribute("configuration-name", CONFIGNAME)
					.addAttribute("id", idManager.getNextId())
					.addAttribute("signLev", "0.05")
					.addAttribute("reqAbsPrc", "50")
					.addAttribute("reqRelPrc", "" + accuracy)
					.addAttribute("batchSize", "200")
					.addAttribute("minBatches", "51")
					.addAttribute("numBMeansCorlTested", "0")
					.addAttribute("bucketSize", "100.0")
					.addAttribute("maxBuckets", "1000")
					.addAttribute("queueSignLev", "0.05")
					.addAttribute("queueReqAbsPrc", "50")
					.addAttribute("queueReqRelPrc", "" + accuracy)
					.addAttribute("queueBatchSize", "200")
					.addAttribute("queueMinBatches", "60")
					.addAttribute("queueNumBMeansCorlTested", "50")
					.addAttribute("queueBucketSize", "100.0")
					.addAttribute("queueMaxBuckets", "1000");
			}
		}
	}
	
	private void addMainConfig(Element net, IdManager idManager) {
		Element mainNet = (Element)net.selectSingleNode("//net");
	
		Element metaAttributes = setupMetaAttributes(mainNet);
		
		metaAttributes.addElement("meta-attribute")
			.addAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
			.addAttribute("configuration-name", CONFIGNAME)
			.addAttribute("xsi:type", "simqpn-configuration")
			.addAttribute("id", idManager.getNextId())
			.addAttribute("output-directory", outputDirectory)
			.addAttribute("ramp-up-length", "" + warmupPeriod)
			.addAttribute("scenario", "1")
			.addAttribute("time-before-initial-heart-beat", "" + timeBeforeInitialHeartBeat)
			.addAttribute("seconds-between-heart-beats", "" + secondsBetweenHeartBeats)
			.addAttribute("verbosity-level", "" + verbosity)
			.addAttribute("stopping-rule", stoppingRule)
			.addAttribute("time-between-stop-checks", "" + timeBetweenStopChecks)
			.addAttribute("seconds-between-stop-checks", "" + secondsBetweenStopChecks)
			.addAttribute("total-run-length", "" + totalRunLength);
	}
	
	public Double getAccuracy() {
		return accuracy;
	}
	
	public Double getWarmupPeriod() {
		return warmupPeriod;
	}
	public Double getTotalRunLength() {
		return totalRunLength;
	}

	public void setAccuracy(Double accuracy) {
		this.accuracy = accuracy;
	}

	public Double getSecondsBetweenStopChecks() {
		return secondsBetweenStopChecks;
	}
	
	public class StatLevel {
		private String placeId;
		private int level;
		
		public StatLevel(String placeId, int level) {
			if (level < 0 || level > 4)
				log.error("The statlevel has to be between 0 and 4!");
			this.placeId = placeId;
			this.level = level;
		}
		
		public String getPlaceId() {
			return placeId;
		}
		
		public int getLevel() {
			return level;
		}
		
		public void setLevel(int level ) {
			if (level < 0 || level > 4)
				log.error("The statlevel has to be between 0 and 4!");
			this.level = level;
		}
	}
}
