/**
 * ==============================================
 *  DQL Adapter for SimQPN
 * ==============================================
 *
 * (c) Copyright 2014-2016, by Simon Eismann, Juergen Walter and other contributors.
 *
 * All rights reserved. This software is made available under the terms of the
 * Eclipse Public License (EPL) v1.0 as published by the Eclipse Foundation
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the Eclipse Public License (EPL)
 * for more details.
 *
 * You should have received a copy of the Eclipse Public License (EPL)
 * along with this software; if not visit http://www.eclipse.org or write to
 * Eclipse Foundation, Inc., 308 SW First Avenue, Suite 110, Portland, 97204 USA
 * Email: license (at) eclipse.org
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */
package tools.descartes.dql.connector.simqpn.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.XMLConstants;

import org.apache.log4j.Logger;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.QName;
import org.dom4j.XPath;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.XMLWriter;

import tools.descartes.dql.connector.simqpn.exception.DQLException;
import tools.descartes.dql.core.engine.util.DQLLogger;

public class NetUtil {

	private static final Logger log = DQLLogger.getLogger(NetUtil.class.getName());
	
	public static List<Element> listQueues(Element model) {
		return queryElements(model, "//queues/queue");
	}

	public static List<Element> listColorReferences(Element place) {
		return queryElements(place, "color-refs/color-ref");
	}

	public static String getNameOfUnderlyingQueue(String placeName, Element net) {
		Element p = NetUtil.getPlaceByName(placeName, net);
		String queueId = p.attributeValue(new QName("queue-ref"));
		if (queueId == null)
			throw new DQLException("That is not a queueing Place");
		String queueName = NetUtil.getNameOfElement(getQueueById(net, queueId));
		return queueName;
	}
	
	public static Element getQueueById(Element model, String id) {
		return element(model, "//queue[@id = '" + id + "']");
	}
	
	private static String getUnderlyingQueueName(Element p, Element net) {
		String qId = p.attributeValue(new QName("queue-ref"));
		if (qId == null)
			return null;
		Element queue = getQueueById(net, qId);
		return getNameOfElement(queue);
	}

	public static String getNameOfElement(Element e) {
		return e.attributeValue(new org.dom4j.QName("name"));
	}

	public static String getIdOfElement(Element e) {
		return e.attributeValue(new org.dom4j.QName("id"));
	}

	public static String getTypeOfPlace(Element place) {
		if (place.getName().equals("probe"))
				return "probe";
		if (place.getName().equals("queue"))
			return "queue";
		String type = place.attributeValue(new org.dom4j.QName("type"));
		if (type == null)
			type = place.attributeValue(new org.dom4j.QName("type",
					new Namespace("xsi", XMLConstants.W3C_XML_SCHEMA_INSTANCE_NS_URI)));
		return type;
	}
	
	public static String getIdOfPlaceByName(String name, Element net) {
		return getIdOfElement(getPlaceByName(name, net));
	}
	
	public static String removeQueueTag(String name) {
		return name.substring(6);
	}
	
	public static Element getProbeByName(String name, Element net) {
		for (Element e: listProbes(net))
			if (getNameOfElement(e).equals(name))
				return e;
		return null;
	}
	
	public static List<Element> listProbes(Element net) {
		return queryElements(net, "./probes/probe");
	}
	
	public static Element getPlaceByName(String name, Element net) {
		for (Element e: listAllPlaces(net))
			if (getNameOfElement(e).equals(name))
				return e;
		return null;
	}
	
	public static List<Element> listAllPlaces(Element model) {
		return queryElements(model, "//place");
	}
	
	public static String getReferencedColorId(Element e) {
		return e.attributeValue(new org.dom4j.QName("color-id"));
	}

	public static int getInitialPopulation(Element e) {
		return Integer.valueOf(e.attributeValue(new org.dom4j.QName("initial-population")));
	}
	
	public static void save(File path, Element doc) {
        XMLWriter writer = null;
        try {
            OutputFormat format = OutputFormat.createPrettyPrint();
            writer = new XMLWriter(new FileOutputStream(path), format);
            writer.write(doc);
        } catch (IOException e) {
            log.error(e.getMessage());
        } finally {
            if (writer != null) {
                try {
                    writer.close();
                } catch (IOException e) {
                    log.error(e.getMessage());
                }
            }
        }
    }

	public static String getIdOfPlaceWithQueue(String queueName, Element net) {
		for (Element place: listPlaces(net))
			if (getUnderlyingQueueName(place, net) != null &&
				getUnderlyingQueueName(place, net).equals(queueName))
				return getIdOfElement(place);
		return null;
	}
	
	public static List<Element> listPlaces(Element net) {
		return queryElements(net, "places/place");
	}
	
	public static List<Element> queryElements(Element elem, String xpath) {
		XPath xpathSelector = DocumentHelper.createXPath(xpath);
		Map<String, String> namespaceUris = new HashMap<String, String>();
		namespaceUris.put("xsi", "http://www.w3.org/2001/XMLSchema-instance");
		xpathSelector.setNamespaceURIs(namespaceUris);
		List<Element> ret = new ArrayList<Element>();
		for (Object o: xpathSelector.selectNodes(elem))
			ret.add((Element)o);
		return ret;
	}

	public static String getIdOfProbeByName(String identifier, Element net) {
		return getIdOfElement(getProbeByName(identifier, net));
	}

	public static Element getProbeById(Element net, String id) {
		for (Element e: listProbes(net))
			if (getIdOfElement(e).equals(id))
				return e;
		return null;
	}
	
	public static List<Element> getColorReferencesForPlace(Element place) {
		List<Element> refs = new LinkedList<Element>();
		for (Object o: place.selectNodes("color-refs/color-ref"))
			refs.add((Element)o);
		return refs;
	}
	
	public static Element element(Element elem, String xpath) {
		XPath xpathSelector = DocumentHelper.createXPath(xpath);
		return (Element)xpathSelector.selectSingleNode(elem);
	}
	
	public static Element getElement(Element elem, String id) {
		return element(elem, "//*[@id = '" + id + "']");
	}
}
